# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
from itertools import product

import numpy as np
import scipy.special as spsp

from .util import ERROR, get_as_list, tagged


class Wavevector:
    def __init__(self):
        pass

    def generate(self, param, size):
        self.dim = len(size)
        steps = get_as_list(param, "ksteps", default=0, extendto=self.dim)
        for d in range(self.dim):
            if steps[d] == 0:
                steps[d] = size[d]//2
        ks = []
        self.nk = 1
        for d in range(self.dim):
            ks.append(list(range(0, size[d] // 2 + 1, steps[d])))
            self.nk *= len(ks[-1])
        self.ks = np.zeros([self.dim, self.nk], dtype=np.int)
        for ik, k in enumerate(product(*ks)):
            self.ks[:, ik] = list(k)

    def load(self, inp):
        if type(inp) is str:
            with open(inp) as f:
                self.load(f)
                return

        self.dim = 0
        state = "waiting"
        for line in inp:
            body = line.split("#")[0].strip()
            if len(body) == 0:
                continue
            if state == "waiting":
                if body.lower() == "dim":
                    state = "dim"
                else:
                    continue
            elif state == "dim":
                self.dim = int(body)
                break
        if not self.dim > 0:
            ERROR("invalid dimension")

        state = "waiting"
        for line in inp:
            body = line.split("#")[0].strip()
            if len(body) == 0:
                continue
            if state == "waiting":
                if body.lower() == "dim":
                    continue
                elif body.lower() == "kpoints":
                    state = "nk"
            elif state == "nk":
                self.nk = int(body)
                self.ks = np.zeros([self.dim, self.nk], dtype=np.int)
                state = "kpoints"
            elif state == "kpoints":
                words = body.split()
                kid = int(words[0])
                self.ks[:, kid] = list(map(int, words[1:]))

    def save(self, out):
        if type(out) is str:
            with codecs.open(out, "w", "utf-8") as f:
                self.save(f)
                return

        out.write("dim\n{0}\n".format(self.dim))
        out.write("\n")

        out.write("kpoints\n")
        out.write("{0}\n".format(self.nk))
        for ik in range(self.nk):
            out.write(str(ik))
            for k in self.ks[:, ik]:
                out.write(" {0}".format(k))
            out.write("\n")

    def write_xml(self, filename, lat):
        if self.dim != lat.dim:
            ERROR("dimension mismatch between wavevector and lattice")
        with codecs.open(filename, "w", "utf-8") as f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write("<WaveVector>\n")
            f.write(tagged("Comment", lat.name))
            f.write(tagged("NumberOfSites", len(lat.sites)))
            f.write(tagged("NumberOfWaveVectors", self.nk))

            f.write("<!-- <RK> [phase(cos)] [phase(sin)] [isite] [kindx] </RK> -->\n")
            for ik in range(self.nk):
                k = self.ks[:, ik]
                for site in lat.sites:
                    sid = site.id
                    coord = site.coord
                    phase = 0
                    for d in range(self.dim):
                        phase += coord[d] * k[d] / float(lat.size[d])
                    c = spsp.cosdg(360 * phase)
                    s = spsp.sindg(360 * phase)
                    f.write(tagged("RK", [c, s, sid, ik]))

            f.write("</WaveVector>\n")


def main():
    import argparse
    import toml

    parser = argparse.ArgumentParser(
        description="Generate wavevector file for DSQSS/DLA",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    parser.add_argument("input", help="Input filename")
    parser.add_argument(
        "-o", "--output", dest="out", default="kpoints.dat", help="Output filename"
    )
    parser.add_argument(
        "-s",
        "--size",
        dest="size",
        default=None,
        help='system size (specified by space separated integers like "4 4"). '
        + "This option ignores the setting in [lattice] section of input TOML file",
    )

    args = parser.parse_args()

    Ls = None
    inp = toml.load(args.input)
    if "lattice" in inp:
        dim = inp["lattice"]["dim"]
        Ls = get_as_list(inp["lattice"], "L", extendto=dim)
    if "kpoints" in inp:
        inp = inp["kpoints"]
    if args.size is not None:
        Ls = list(map(int, args.size.split()))
    if Ls is None:
        ERROR("system size is not specified")
    wv = Wavevector()
    wv.generate(inp, Ls)
    wv.save(args.out)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# from .hypercubic import HyperCubicLattice
import codecs

from .lattice import Lattice
from .lattice_factory import honeycomb, hypercubic, kagome, triangular
from .util import ERROR


def std_lattice(param, ret_dict=False):
    if "lattice" in param and not isinstance(param["lattice"], str):
        param = param["lattice"]
    if "unitcell" in param:
        latticedict = param
    else:
        latname = param["lattice"].lower()
        if latname == "hypercubic":
            latticedict = hypercubic.generate(param)
        elif latname == "triangular":
            latticedict = triangular.generate(param)
        elif latname == "honeycomb":
            latticedict = honeycomb.generate(param)
        elif latname == "kagome":
            latticedict = kagome.generate(param)
        else:
            ERROR(
                'Unknown lattice: param["lattice"] = {0}'.format(
                    param["lattice"].lower()
                )
            )

    lat = Lattice()
    lat.load_dict(latticedict)
    if ret_dict:
        return lat, latticedict
    else:
        return lat


def main():
    import argparse
    import toml

    parser = argparse.ArgumentParser(
        description="Generate lattice file for DSQSS/DLA",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    parser.add_argument("input", help="Input filename")
    parser.add_argument(
        "-o", "--output", dest="out", default="lattice.dat", help="Output data filename"
    )
    parser.add_argument(
        "-t", "--toml", dest="toml", default="", help="Output TOML filename"
    )
    parser.add_argument(
        "-g", "--gnuplot", dest="gnuplot", default="", help="Output Gnuplot filename"
    )

    args = parser.parse_args()
    istoml = True
    try:
        inp = toml.load(args.input)
    except toml.TomlDecodeError:
        istoml = False
        print("Input file seems not to be TOML file.")

    if istoml:
        lat, lattice_dict = std_lattice(inp, True)
    else:
        lat = Lattice(args.input)

    if len(args.out) > 0:
        lat.save_dat(args.out)
    if len(args.toml) > 0:
        if istoml:
            with codecs.open(args.toml, "w", "utf-8") as f:
                toml.dump(lattice_dict, f)
        else:
            ERROR("Cannot generate a lattice TOML file from a lattice data file.")

    if len(args.gnuplot) > 0:
        lat.write_gnuplot(args.gnuplot)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import argparse
import os.path
import subprocess
import sys
import warnings

from .read_keyvalues import read_keyvalues
from .util import ERROR


class param:
    def __init__(self):
        self.param_dict = {
            "beta": 10.0,
            "npre": 1000,
            "ntherm": 1000,
            "ndecor": 1000,
            "nmcs": 1000,
            "nset": 10,
            "seed": 31415,
            "nvermax": 10000000,
            "algfile": "algorithm.xml",
            "latfile": "lattice.xml",
            "outfile": "sample.log",
            "sfinpfile": "",
            "cfinpfile": "",
            "ckinpfile": "",
            "sfoutfile": "sf.dat",
            "cfoutfile": "cf.dat",
            "ckoutfile": "ck.dat",
        }

        self.param_dict_pmwa = {
            "runtype": 0,
            "cb": 0,
            "nwormax": 1000,
            "step_x": 1,
            "step_k": 1,
            # Hamiltonian
            "t": 0.0,
            "u": 0.0,
            "v": 0.0,
            "mu": 0.0,
            "g": 0.0,
        }

        self.info_dict = {
            # System
            "solver": "DLA",
            "model_type": "spin",
        }

        self.info_dict_dla = {
            # Hamiltonian
            "m": 1,
            "j": 1.0,
            "f": 0.0,
            "v": 0.0,
            "u": 0.0,
            # Lattice
            "lattice_type": "square",
            "d": 1,
            "l": "8",
            "beta": 10.0,
            # sf,cf
            "ntau": 100,
            "ntau_cutoff": 100,
            "ktype": 0,
        }

        self.info_dict_pmwa = {
            # Lattice
            "lattice_type": "square",
            "d": 1,
            "l": 8,
            "beta": 10.0,
            "nldiv": 1,
            "nbdiv": 1,
            "nfield": 0,  # to be deleted
            # additonal information
            "nmax": 1,  # fixed 1
        }


class info:
    def __init__(self, args):
        self.args = args
        self.dict_lattice = {"square": "lattgene_C", "triangular": "lattgene_T"}
        self.dict_hamgen = {"spin": "hamgen_H", "boson": "hamgen_B"}
        self.dict_solver = {"DLA": "dla_", "PMWA": "pmwa_"}
        self.prm = param()
        self.gendir = args.gendir
        if self.gendir is None:
            self.gendir = os.path.dirname(os.path.abspath(sys.argv[0]))
            print("use file generators in {0}".format(self.gendir))
        self._get_info(args.input)
        self._check_condition()

    def get_solver(self):
        if self.prm.info_dict["model_type"] == "spin":
            model_label = "H"
        else:
            model_label = "B"
        return self.dict_solver[self.prm.info_dict["solver"]] + model_label

    def _check_condition(self):
        if self.prm.info_dict["solver"] == "PMWA":
            if self.prm.info_dict["lattice_type"] != "square":
                raise RuntimeError("For PMWA, square type lattice is only used.")

    def _get_lattgene(self):
        lattgene_name = self.dict_lattice[self.prm.info_dict["lattice_type"]]
        if self.prm.info_dict["solver"] == "PMWA":
            lattgene_name = "lattgene_P"
        return os.path.join(self.gendir, lattgene_name)

    def _get_info(self, inputfile):
        tmp_dict = read_keyvalues(inputfile)
        """
        if inputfile is sys.stdin:
            print('Waiting for standard input...')
        data_list = inputfile.readlines()
        tmp_dict={}
        # get key and value
        for data in data_list:
            data.strip()
            d_re = re.search("(.*)#(.*)", data)
            if d_re is not None:
                data = d_re.group(1)
            pattern = "(.*)=(.*)"
            if re.search(pattern,data) is not None:
                d_re = re.search(pattern,data)
                key = d_re.group(1).strip().lower()
                tmp_dict[key] = d_re.group(2).strip()
        """

        if tmp_dict["solver"] == "DLA":
            ERROR("Please use dla_pre instead of pmwa_pre")
        else:
            self.prm.info_dict.update(self.prm.info_dict_pmwa)
            self.prm.param_dict.update(self.prm.param_dict_pmwa)
            # if oldbeta exits, add param_dict
            if tmp_dict.get("oldbeta") is not None:
                self.prm.param_dict["oldbeta"] = float(tmp_dict.get("oldbeta"))

            # modify parameters for spin model
            if ("model_type" in tmp_dict) is False:
                tmp_dict["model_type"] = self.prm.info_dict["model_type"]

            if tmp_dict["model_type"] == "spin":
                # assume that the lattice is bipartite
                # this makes that t is always positive (by unitary transform)
                if tmp_dict.get("jxy") is not None:
                    tmp_dict["jxy"] = abs(float(tmp_dict["jxy"]))
                self._replace_keyword(tmp_dict, "t", "jxy")
                self._replace_keyword(tmp_dict, "v", "jz")
                self._replace_keyword(tmp_dict, "mu", "h")
                self._replace_keyword(tmp_dict, "g", "gamma")
                tmp_dict["g"] = float(tmp_dict["g"]) * 0.5
            else:
                self._replace_keyword(tmp_dict, "g", "gamma")

        self.param_list = list(self.prm.param_dict)
        self.param_list.sort()

        for key in tmp_dict.keys():
            invalid = True
            if key in self.prm.info_dict:
                self.prm.info_dict[key] = tmp_dict[key]
                invalid = False
            if key in self.prm.param_dict:
                self.prm.param_dict[key] = tmp_dict[key]
                invalid = False
            if invalid:
                warnings.warn(
                    'Unknown keyword " ' + key + '" is found (ignored).', Warning
                )

    def _replace_keyword(self, _tmp_dict, origin, new):
        if _tmp_dict.get(new) is not None:
            _tmp_dict[origin] = _tmp_dict[new]
            del _tmp_dict[new]

    def _get_latticeinfo(self):
        linfo = self.prm.info_dict["l"]
        if not isinstance(linfo, list):
            linfo = [linfo]
        if self.prm.info_dict["lattice_type"] == "triangular":
            LatticeInfo = []
        else:
            LatticeInfo = [self.prm.info_dict["d"]]
        for lattice in linfo:
            LatticeInfo.append(lattice)
        return LatticeInfo

    def make_xml(self):
        LatticeInfo = self._get_latticeinfo()
        genpath = self._get_lattgene()
        if self.prm.info_dict["solver"] == "PMWA":
            if not os.path.exists(genpath):
                raise RuntimeError(
                    "{0} is not found: use correct --generator_dir.".format(genpath)
                )
            cmd = [genpath, "-o", self.prm.param_dict["latfile"]]
            for l in LatticeInfo:
                cmd.append(str(l))
            cmd.append(str(self.prm.info_dict["beta"]))

            cmd += [
                str(self.prm.info_dict["nldiv"]),
                str(self.prm.info_dict["nbdiv"]),
                str(self.prm.info_dict["nfield"]),
            ]
            print("\n " + " ".join(cmd))
            subprocess.call(cmd)
        else:
            msg = "pmwa_pre treats only PMWA."
            sys.exit(1)

    def print_init_param(self):
        with open(self.args.pfile, mode="w") as f:
            for param in self.param_list:
                f.write(param + " = " + str(self.prm.param_dict[param]) + "\n")


def main():
    parser = argparse.ArgumentParser(
        description="Generate input files for dsqss", add_help=True
    )

    parser.add_argument(
        "-g",
        "--generator-dir",
        dest="gendir",
        help="Path to directory including generators (lattgene_C, hamgen_B, etc)",
    )
    parser.add_argument(
        "-p", "--paramfile", dest="pfile", default="param.in", help="Parameter file"
    )
    parser.add_argument(
        "--hamfile", dest="hfile", default="hamiltonian.xml", help="Hamiltonian file"
    )
    parser.add_argument("input", type=argparse.FileType("r"), help="Input file")
    parser.add_argument(
        "-s",
        "--stdin",
        dest="stdin",
        action="store_true",
        help="Read from standard input (overrides --input)",
    )

    args = parser.parse_args()
    if args.stdin:
        args.input = sys.stdin
    if args.input is None:
        raise RuntimeError(
            "Neither --input (read from file) nor --stdin (read from stdin) option is used."
        )
    calcinfo = info(args)
    calcinfo.make_xml()
    # Make initial parameter
    calcinfo.print_init_param()
    solver_name = calcinfo.get_solver()
    print("Please type: " + solver_name + " " + args.pfile)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .bosehubbard import BoseHubbard_hamiltonian
from .util import ERROR
from .xxz import XXZ_hamiltonian
from .hamiltonian import Hamiltonian


def std_model(param):
    if "hamiltonian" in param:
        param = param["hamiltonian"]
    if "model" in param:
        if param["model"].lower() == "spin":
            ham = XXZ_hamiltonian(param)
        elif param["model"].lower() == "boson":
            ham = BoseHubbard_hamiltonian(param)
        else:
            ERROR('Unknown model: param["model"] = {0}'.format(param["model"].lower()))
    else:
        ham = Hamiltonian(param)

    return ham


def main():
    import argparse
    import toml

    parser = argparse.ArgumentParser(
        description="Generate hamiltonian TOML file for DSQSS/DLA",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    parser.add_argument("input", help="Input filename")
    parser.add_argument(
        "-o", "--output", dest="out", default="hamiltonian.toml", help="Output filename"
    )
    args = parser.parse_args()

    inp = toml.load(args.input)
    if "hamiltonian" in inp:
        inp = inp["hamiltonian"]
    ham = std_model(inp)
    ham.write_toml(args.out)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import codecs
import sys
from itertools import chain

import toml

import numpy as np

from .util import ERROR, WARN, coord2index, get_as_list, index2coord, tagged


class Site:
    def __init__(self, site_id, site_type, coordinate):
        self.id = site_id
        self.stype = site_type
        self.coord = coordinate
        self.z = 0

    def __str__(self):
        s = "{0} {1}".format(self.id, self.stype)
        for c in self.coord:
            s += " {0}".format(c)
        return s


class Interaction:
    def __init__(self, int_id, int_type, nbody, site_indices, edge_flag, direction):
        self.id = int_id
        self.itype = int_type
        self.itype_org = self.itype
        self.nbody = nbody
        self.sites = site_indices
        self.edge = edge_flag
        self.dir = direction
        self.vtype = -1

    def __str__(self):
        s = "{0} {1} {2}".format(self.id, self.itype, self.nbody)
        edgeflag = 1 if self.edge >= 0 else 0
        for site in self.sites:
            s += " {0}".format(site)
        s += " {0} {1}".format(edgeflag, self.dir)
        return s


class Vertex:
    def __init__(self, v_id, int_type, site_types, zs):
        self.v_id = v_id
        self.int_type = int_type
        self.site_types = site_types
        self.zs = zs


class Lattice:
    def __init__(self, inp=None):
        if inp is not None:
            self.load(inp)

    def load(self, inp):
        if type(inp) is dict:
            self.load_dict(inp)
        elif inp.endswith(".toml"):
            self.load_dict(toml.load(inp))
        else:
            self.load_dat(inp)

    def load_dict(self, param):
        if "lattice" in param and not isinstance(param["lattice"], str):
            return self.load_dict(param["lattice"])

        parameter = param["parameter"]
        self.name = parameter["name"]
        self.dim = parameter["dim"]
        self.size = get_as_list(parameter, "L", extendto=self.dim)
        self.bc = get_as_list(parameter, "bc", default=True, extendto=self.dim)
        ncells = np.product(self.size)
        self.latvec = np.array(parameter["basis"], dtype=float).transpose()

        unitcell = param["unitcell"]
        nsites_cell = len(unitcell["sites"])
        nbonds_cell = len(unitcell["bonds"])

        localsitecoords = []
        for site in unitcell["sites"]:
            localsitecoords.append(np.array(site["coord"], dtype=float))

        directionmap = {}
        directions = []
        self.dirs = []
        for ib, bond in enumerate(unitcell["bonds"]):
            if "offset" in bond["source"]:
                if not all(bond["source"]["offset"] == 0):
                    ERROR("bonds['source'] has nonzero offset")

            offset = np.array(
                get_as_list(bond["target"], "offset", default=0.0, extendto=self.dim),
                dtype=float,
            )
            offset += localsitecoords[bond["target"]["siteid"]]
            offset -= localsitecoords[bond["source"]["siteid"]]
            key = tuple(offset)
            if key not in directionmap:
                self.dirs.append(offset)
                directionmap[key] = len(directionmap)
            directions.append(directionmap[key])
        self.ndir = len(self.dirs)

        def bondsites(center_cellcoord, bond):
            ret = []
            offset = center_cellcoord + np.array(
                get_as_list(bond["target"], "offset", default=0, extendto=self.dim)
            )
            edge = 0
            for dim in range(self.dim):
                if offset[dim] < 0:
                    if self.bc[dim]:
                        edge = 1
                    else:
                        return False, 0
                elif offset[dim] >= self.size[dim]:
                    if self.bc[dim]:
                        edge = 1
                    else:
                        return False, 0

            for site in (bond["source"], bond["target"]):
                cell_coord = center_cellcoord + np.array(
                    get_as_list(site, "offset", default=0, extendto=self.dim)
                )
                for dim in range(self.dim):
                    if cell_coord[dim] < 0:
                        cell_coord[dim] += self.size[dim]
                    elif cell_coord[dim] >= self.size[dim]:
                        cell_coord[dim] -= self.size[dim]
                icell = coord2index(cell_coord, self.size)
                sid = site["siteid"] + icell * nsites_cell
                ret.append(sid)
            if ret[0] == ret[1]:
                WARN("selfloop {0}=>{0} appears.".format(ret[0]))
            return ret, edge

        self.sites = []
        self.ints = []
        bid = 0
        for icell in range(ncells):
            cell_coord = np.array(index2coord(icell, self.size))
            for lid, site in enumerate(unitcell["sites"]):
                sid = site["siteid"] + icell * nsites_cell
                coord = cell_coord + np.array(site["coord"], dtype=float)
                S = Site(sid, site["type"], coord)
                self.sites.append(S)
            for ib, bond in enumerate(unitcell["bonds"]):
                nbody = 2
                sites, edge = bondsites(cell_coord, bond)
                if not sites:
                    continue
                INT = Interaction(bid, bond["type"], nbody, sites, edge, directions[ib])
                self.ints.append(INT)
                bid += 1
        self.nsites = len(self.sites)
        self.nints = len(self.ints)
        self.update()

    def save_dat(self, out):
        if type(out) is str:
            with codecs.open(out, "w", "utf-8") as f:
                self.save_dat(f)
                return

        out.write("name\n")
        out.write("{0}\n".format(self.name))
        out.write("\n")

        out.write("lattice\n")
        out.write("{0} # dim\n".format(self.dim))
        for x in self.size:
            out.write("{0} ".format(x))
        out.write("# size\n")
        for x in self.bc:
            b = 1 if x else 0
            out.write("{0} ".format(b))
        out.write("# 0:open boundary, 1:periodic boundary\n")
        for d in range(self.dim):
            out.write("{0} ".format(d))
            for x in self.latvec[:, d]:
                out.write("{0} ".format(x))
            out.write("# latvec_{0}\n".format(d))
        out.write("\n")

        out.write("directions\n")
        out.write("{0} # ndirections\n".format(self.ndir))
        out.write("# id, coords...\n")
        for i, dir in enumerate(self.dirs):
            out.write("{0} ".format(i))
            for x in dir:
                out.write("{0} ".format(x))
            out.write("\n")
        out.write("\n")

        out.write("sites\n")
        out.write("{0} # nsites\n".format(self.nsites))
        out.write("# id, type, coord...\n")
        for i, site in enumerate(self.sites):
            out.write("{0}\n".format(site))
        out.write("\n")

        out.write("interactions\n")
        out.write("{0} # nints\n".format(self.nints))
        out.write("# id, type, nbody, sites..., edge_flag, direction\n")
        for i, inter in enumerate(self.ints):
            out.write("{0}\n".format(inter))

    def load_dat(self, inp):
        if type(inp) is str:
            with open(inp) as f:
                self.load_dat(f)
                return

        state = "waiting"
        count = 0
        self.dim = 0
        self.nsites = 0
        self.nints = 0

        # search for 'lattice' because others requires dimension
        for line in inp:
            body = line.split("#")[0].strip()
            if len(body) == 0:
                continue
            if state == "waiting" and body.lower() == "lattice":
                state = "dim"
            elif state == "dim":
                self.dim = int(body)
                self.size = None
                self.bc = [None] * self.dim
                self.latvec = np.eye(self.dim)
                state = "size"
            elif state == "size":
                self.size = list(map(int, body.split()))
                if len(self.size) != self.dim:
                    if len(self.size) < self.dim:
                        ERROR("too few elements ({0})".format(body))
                    else:
                        ERROR("too many elements ({0})".format(body))
                state = "bc"
            elif state == "bc":
                elem = body.split()
                if len(elem) != self.dim:
                    if len(elem) < self.dim:
                        ERROR("too few elements ({0})".format(body))
                    else:
                        ERROR("too many elements ({0})".format(body))
                self.bc[:] = list(map(int, elem))
                state = "latvec"
            elif state == "latvec":
                self.load_latvec(body, count)
                count += 1
                if count == self.dim:
                    count = 0
                    break

        # search for others
        inp.seek(0)
        count = 0
        state = "waiting"
        for line in inp:
            body = line.split("#")[0].strip()
            if len(body) == 0:
                continue

            if state == "waiting":
                if body.lower() == "name":
                    state = "name"
                elif body.lower() == "directions":
                    state = "ndir"
                elif body.lower() == "sites":
                    state = "nsites"
                elif body.lower() == "interactions":
                    state = "nints"
            elif state == "name":
                self.name = body
                state = "waiting"
            elif state == "ndir":
                self.ndir = int(body)
                self.dirs = [None for i in range(self.ndir)]
                state = "dirs"
                count = 0
            elif state == "dirs":
                self.load_direction(body, count)
                count += 1
                if count == self.ndir:
                    count = 0
                    state = "waiting"
            elif state == "nsites":
                self.nsites = int(body)
                state = "sites"
                self.sites = [None for i in range(self.nsites)]
            elif state == "sites":
                self.load_site(body, count)
                count += 1
                if count == self.nsites:
                    state = "waiting"
                    count = 0
            elif state == "nints":
                self.nints = int(body)
                state = "ints"
                self.ints = [None for i in range(self.nints)]
            elif state == "ints":
                self.load_int(body, count)
                count += 1
                if count == self.nints:
                    state = "waiting"
                    count = 0

        # end of for line in inp:

        self.check_all()
        self.update()

    def load_latvec(self, body, count):
        elem = body.split()
        if len(elem) != self.dim + 1:
            if len(elem) < self.dim + 1:
                ERROR("too few elements ({0})".format(body))
            else:
                ERROR("too many elements ({0})".format(body))
        self.latvec[:, int(elem[0])] = list(map(float, elem[1:]))

    def load_direction(self, body, count):
        elem = body.split()
        if len(elem) != self.dim + 1:
            if len(elem) < self.dim + 1:
                ERROR("too few elements ({0})".format(body))
            else:
                ERROR("too many elements ({0})".format(body))
        self.dirs[int(elem[0])] = list(map(float, elem[1:]))

    def load_site(self, body, count):
        elem = body.split()
        if len(elem) != self.dim + 2:
            if len(elem) < self.dim + 2:
                ERROR("too few elements ({0})".format(body))
            else:
                ERROR("too many elements ({0})".format(body))
        self.sites[int(elem[0])] = Site(
            site_id=int(elem[0]),
            site_type=int(elem[1]),
            coordinate=list(map(float, elem[2:])),
        )

    def load_int(self, body, count):
        elem = body.split()
        nbody = int(elem[2])
        if len(elem) != nbody + 5:
            if len(elem) < nbody + 5:
                ERROR("too few elements ({0})".format(body))
            else:
                ERROR("too many elements ({0})".format(body))
        self.ints[int(elem[0])] = Interaction(
            int_id=int(elem[0]),
            int_type=int(elem[1]),
            nbody=nbody,
            site_indices=list(map(int, elem[3 : (3 + nbody)])),
            edge_flag=int(elem[-2]),
            direction=int(elem[-1]),
        )

    def check_all(self):
        self.check_latvec()
        self.check_dirs()
        self.check_sites()
        self.check_ints()

    def check_latvec(self):
        invalid = False
        for i in range(self.dim):
            if self.latvec[i] is None:
                ERROR("lattice vector {0} is not defined".format(i), True)
                invalid = True
        if invalid:
            sys.exit(1)

    def check_dirs(self):
        invalid = False
        for i in range(self.ndir):
            if self.dirs[i] is None:
                ERROR("bond direction {0} is not defined".format(i), True)
                invalid = True
        if invalid:
            sys.exit(1)

    def check_sites(self):
        invalid = False
        for i in range(self.nsites):
            if self.sites[i] is None:
                ERROR("site {0} is not defined".format(i), True)
                invalid = True
        if invalid:
            sys.exit(1)

    def check_ints(self):
        invalid = False
        for i in range(self.nints):
            if self.ints[i] is None:
                ERROR("interaction {0} is not defined".format(i), True)
                invalid = True
        if invalid:
            sys.exit(1)

    def update(self):
        self.nstypes = 0
        for s in self.sites:
            s.z = 0
            self.nstypes = max(self.nstypes, s.stype)
        self.nstypes += 1

        self.nedges = 0
        for inter in self.ints:
            if inter.edge == 1:
                inter.edge = self.nedges
                self.nedges += 1
                pass
            else:
                inter.edge = -1
                pass
            for s in inter.sites:
                self.sites[s].z += 1

        self.nitypes = 0
        self.nvtypes = 0
        self.vertices = []
        vmap = {}
        for inter in self.ints:
            itype = inter.itype
            self.nitypes = max(self.nitypes, itype)
            stypes = tuple(self.sites[s].stype for s in inter.sites)
            zs = tuple(self.sites[s].z for s in inter.sites)
            if (itype, stypes, zs) in vmap:
                inter.itype = vmap[(itype, stypes, zs)]
            else:
                inter.itype = vmap[(itype, stypes, zs)] = self.nvtypes
                self.vertices.append(Vertex(self.nvtypes, itype, stypes, zs))
                self.nvtypes += 1
        self.nitypes += 1

    def write_xml(self, filename):
        with codecs.open(filename, "w", "utf_8") as f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write("<LATTICE>\n")

            f.write(tagged("Comment", self.name))

            f.write(tagged("Dimension", self.dim))
            f.write(tagged("LinearSize", self.size))
            f.write(tagged("BoundaryCondition", self.bc))
            f.write(tagged("NumberOfSites", self.nsites))
            f.write(tagged("NumberOfInteractions", self.nints))
            f.write(tagged("NumberOfSiteTypes", self.nstypes))
            f.write(tagged("NumberOfBondDirections", len(self.dirs)))
            f.write(tagged("NumberOfInteractionTypes", self.nvtypes))
            f.write(tagged("NumberOfEdgeInteractions", self.nedges))
            f.write("\n")

            for d in range(self.dim):
                f.write(tagged("Basis", self.latvec[:, d]))
            f.write("\n")

            f.write("<!-- <S> site_index site_type </S> -->\n")
            for i, site in enumerate(self.sites):
                f.write(tagged("S", (i, site.stype)))
            f.write("\n")

            f.write("<!-- <SiteCoordinate> site_index coord... </SiteCoordinate> -->\n")
            for i, site in enumerate(self.sites):
                f.write(tagged("SiteCoordinate", chain([i], site.coord)))
            f.write("\n")

            f.write(
                (
                    "<!-- <I> int_index int_type nbody site_index... "
                    "edge_index direction_index </I> -->\n"
                )
            )
            for i, interaction in enumerate(self.ints):
                f.write(
                    tagged(
                        "I",
                        chain(
                            [i, interaction.itype, len(interaction.sites)],
                            interaction.sites,
                            [interaction.edge, interaction.dir],
                        ),
                    )
                )
            f.write("\n")

            f.write("<!-- <Direction> direction_index direction... </Direction> -->\n")
            for i, bond in enumerate(self.dirs):
                f.write(tagged("Direction", chain([i], bond)))

            f.write("</LATTICE>\n")

    def write_gnuplot(self, filename):
        with open(filename, "w") as f:
            for st in range(self.nstypes):
                f.write("$SITES_{0} << EOD\n".format(st))
                for site in self.sites:
                    if site.stype != st:
                        continue
                    v = np.dot(self.latvec, site.coord)
                    for x in v:
                        f.write("{0} ".format(x))
                    f.write("{0}\n".format(site.stype))
                f.write("EOD\n")

            for bt in range(self.nitypes):
                f.write("$BONDS_{0} << EOD\n".format(bt))
                for bond in self.ints:
                    if bond.nbody != 2:
                        continue
                    if bond.itype_org != bt:
                        continue
                    v = np.dot(self.latvec, np.array(self.sites[bond.sites[0]].coord, dtype=float))
                    for x in v:
                        f.write("{0} ".format(x))
                    f.write("\n")
                    v += np.dot(self.latvec, np.array(self.dirs[bond.dir], dtype=float))
                    for x in v:
                        f.write("{0} ".format(x))
                    f.write("\n\n")
                f.write("EOD\n")

            f.write("plot ")
            for st in range(self.nstypes):
                f.write('$SITES_{0} w p pt {1} ps 2 t "" , \\\n'.format(st, st + 4))
            for bt in range(self.nitypes):
                f.write('$BONDS_{0} w l lw 2 lt {1} t "" , \\\n'.format(bt, bt + 1))
            f.write('\n')
            f.write('pause -1\n')



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class Result:
    def __init__(self, mean, err):
        self.mean = mean
        self.err = err

    def to_str(self, delim="+/-"):
        return "{}{}{}".format(self.mean, delim, self.err)


class Results:
    def __init__(self, resfile):
        self.result = {}
        with open(resfile) as f:
            for line in f:
                if not line.startswith("R"):
                    continue
                words = line.split()
                R = Result(float(words[3]), float(words[4]))
                self.result[words[1]] = R

    def to_str(self, names, delim=" ", delim_data=" "):
        if type(names) is str:
            return self.result[names].to_str(delim)
        ss = [self.result[name].to_str(delim) for name in names]
        return delim_data.join(ss)



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from math import sqrt

from .hamiltonian import Hamiltonian, Interaction, Site, append_matelem
from .util import extend_list, get_as_list


def creator_boson(n):
    return sqrt(n + 1.0)


def annihilator_boson(n):
    return sqrt(n)


class BoseSite(Site):
    def __init__(self, id, M, U, mu):
        """
        M: max n
        U: n_i (n_i - 1)/2
        mu: -n_i
        """
        M = int(M)
        NX = M + 1

        values = []
        sources = {}
        elements = {}
        for n in range(NX):
            value = -mu * n + 0.5 * U * n * (n - 1)
            values.append(n)
            append_matelem(elements, state=n, value=value)
            if n > 0:
                # annihilator
                append_matelem(
                    sources, istate=n, fstate=n - 1, value=annihilator_boson(n)
                )
            if n < M:
                # creator
                value = creator_boson(n)
                append_matelem(sources, istate=n, fstate=n + 1, value=value)
        super(BoseSite, self).__init__(id=id, N=NX, values=values,
                                       elements=elements, sources=sources)


class BoseBond(Interaction):
    def __init__(self, id, M, t, V):
        """
        M: max N
        t: c_1 a_2 + c_2 a_1
        V: n_1 n_2
        """

        nbody = 2
        nx = M + 1
        Ns = [nx, nx]

        N = [1.0 * n for n in range(nx)]
        c = [creator_boson(n) for n in N]
        c[-1] = 0.0
        a = [annihilator_boson(n) for n in N]
        elements = {}
        for i in range(nx):
            for j in range(nx):
                # diagonal
                w = V * i * j
                if w != 0.0:
                    append_matelem(elements, state=[i, j], value=w)
                # offdiagonal
                w = -t * c[i] * a[j]
                if w != 0.0:
                    append_matelem(
                        elements, istate=[i, j], fstate=[i + 1, j - 1], value=w
                    )
                w = -t * a[i] * c[j]
                if w != 0.0:
                    append_matelem(
                        elements, istate=[i, j], fstate=[i - 1, j + 1], value=w
                    )
        super(BoseBond, self).__init__(id=id, nbody=nbody,
                                       Ns=Ns, elements=elements)


class BoseHubbard_hamiltonian(Hamiltonian):
    def __init__(self, param):
        M = param["M"]
        Us = get_as_list(param, "U", 0.0)
        mus = get_as_list(param, "mu", 0.0)
        nstypes = max(len(Us), len(mus))
        extend_list(Us, nstypes)
        extend_list(mus, nstypes)
        self.sites = [BoseSite(i, M, U, mu)
                      for i, (U, mu) in enumerate(zip(Us, mus))]

        ts = get_as_list(param, "t", 0.0)
        Vs = get_as_list(param, "V", 0.0)
        nitypes = max(len(ts), len(Vs))
        extend_list(ts, nitypes)
        extend_list(Vs, nitypes)
        self.interactions = [
            BoseBond(i, M, t, V) for i, (t, V) in enumerate(zip(ts, Vs))
        ]
        self.name = "Bose-Hubbard model"



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from math import sqrt

from .hamiltonian import Hamiltonian, Interaction, Site, append_matelem
from .util import extend_list, get_as_list


def Splus(S, m):
    return sqrt((S - m) * (S + m + 1.0))


def Sminus(S, m):
    return sqrt((S + m) * (S - m + 1.0))


class SpinSite(Site):
    def __init__(self, id, M, D, h):
        """
        M: 2S+1
        D: Sz^2
        h: -Sz
        """
        S = 0.5 * M
        NX = M + 1
        values= []
        elements = {}
        sources = {}
        for st in range(NX):
            m = st - 0.5 * M
            values.append(m)
            append_matelem(elements, state=st, value=-h * m + D * m * m)
            if st > 0:
                # annihilator
                append_matelem(
                    sources, istate=st, fstate=st - 1, value=0.5 * Sminus(S, m)
                )
            if st < M:
                # creator
                append_matelem(
                    sources, istate=st, fstate=st + 1, value=0.5 * Splus(S, m)
                )
        super(SpinSite, self).__init__(id=id, N=NX, values=values,
                                       elements=elements, sources=sources)


class XXZBond(Interaction):
    def __init__(self, id, M, z, x):
        """
        M: 2S+1
        z: Sz_1 Sz_2
        x: Sx_1 Sx_2 + Sy_1 Sy_2
        """

        nbody = 2
        nx = M + 1
        Ns = [nx, nx]
        S = 0.5 * M
        Sz = [i - 0.5 * M for i in range(nx)]
        Sp = [Splus(S, m) for m in Sz]
        Sm = [Sminus(S, m) for m in Sz]
        elements = {}
        for i in range(nx):
            for j in range(nx):
                # diagonal
                w = -z * Sz[i] * Sz[j]
                if w != 0.0:
                    append_matelem(elements, state=[i, j], value=w)

                # offdiagnal
                w = -0.5 * x * Sp[i] * Sm[j]
                if w != 0.0:
                    append_matelem(
                        elements, istate=[i, j], fstate=[i + 1, j - 1], value=w
                    )
                w = -0.5 * x * Sm[i] * Sp[j]
                if w != 0.0:
                    append_matelem(
                        elements, istate=[i, j], fstate=[i - 1, j + 1], value=w
                    )
        super(XXZBond, self).__init__(id=id, nbody=nbody, Ns=Ns, elements=elements)


class XXZ_hamiltonian(Hamiltonian):
    def __init__(self, param):
        M = param["M"]
        Ds = get_as_list(param, "D", 0.0)
        hs = get_as_list(param, "h", 0.0)
        nstypes = max(len(Ds), len(hs))
        extend_list(Ds, nstypes)
        extend_list(hs, nstypes)
        self.sites = [SpinSite(i, M, D, h) for i, (D, h) in enumerate(zip(Ds, hs))]

        Jzs = get_as_list(param, "Jz", 0.0)
        Jxys = get_as_list(param, "Jxy", 0.0)
        nitypes = max(len(Jzs), len(Jxys))
        extend_list(Jzs, nitypes)
        extend_list(Jxys, nitypes)
        self.interactions = [
            XXZBond(i, M, Jz, Jx) for i, (Jz, Jx) in enumerate(zip(Jzs, Jxys))
        ]

        self.name = "S={0}/2 XXZ model".format(M)



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import itertools
from copy import deepcopy

import toml

from .lattice import Lattice
from .util import tagged


def keystate(istate, fstate):
    if type(istate) is not tuple:
        if type(istate) is not list:
            istate = [istate]
        ti = tuple(istate)
    else:
        ti = istate

    if type(fstate) is not tuple:
        if type(fstate) is not list:
            fstate = [fstate]
        tf = tuple(fstate)
    else:
        tf = fstate
    return (ti, tf)


def append_matelem(
    matelems, state=None, istate=None, fstate=None, value=None, param=None
):
    if param is None:
        if istate is not None:
            matelems[keystate(istate, fstate)] = value
        else:
            matelems[keystate(state, state)] = value
    else:
        if "istate" in param:
            matelems[keystate(param["istate"], param["fstate"])] = param["value"]
        else:
            matelems[keystate(param["state"], param["state"])] = param["value"]
    return matelems


def matelems_todict(matelems, keepzero=False):
    return [
        {"istate": list(key[0]), "fstate": list(key[1]), "value": value}
        for key, value in matelems.items()
        if keepzero or value != 0.0
    ]


class Site(object):
    def __init__(self, param=None,
                 id=None, N=None, values=None,
                 elements=None, sources=None):
        if param is not None:
            self.id = param["type"]
            self.N = param["N"]
            self.values = param["values"]
            self.elements = {}
            for x in param["elements"]:
                append_matelem(self.elements, param=x)
            self.sources = {}
            for x in param["sources"]:
                append_matelem(self.sources, param=x)
        else:
            self.id = id
            self.N = N
            self.values = values
            self.elements = elements
            self.sources = sources

    def to_dict(self):
        return {
            "type": self.id,
            "N": self.N,
            "values": self.values,
            "elements": matelems_todict(self.elements),
            "sources": matelems_todict(self.sources),
        }


class Interaction(object):
    def __init__(self, param=None, id=None, nbody=None, Ns=None, elements=None):
        if param is not None:
            self.id = param["type"]
            self.nbody = param["nbody"]
            self.Ns = param["N"]
            self.elements = {}
            for x in param["elements"]:
                append_matelem(self.elements, param=x)
        else:
            self.id = id
            self.nbody = nbody
            self.Ns = Ns
            self.elements = elements

    def to_dict(self):
        return {
            "type": self.id,
            "nbody": self.nbody,
            "N": self.Ns,
            "elements": matelems_todict(self.elements),
        }


class IndeedInteraction(object):
    def __init__(self, sites, ints, v):
        inter = ints[v.int_type]
        zs = v.zs

        self.itype = v.v_id
        self.stypes = v.site_types
        self.nbody = len(self.stypes)
        self.elements = deepcopy(inter.elements)

        site_elems = [deepcopy(sites[stype].elements) for stype in self.stypes]

        for state in itertools.product(*map(range, inter.Ns)):
            val = 0.0
            for st, els, z in zip(state, site_elems, zs):
                val += els.get(keystate(st, st), 0.0) / z
            key = keystate(state, state)
            if key in self.elements:
                self.elements[key] += val
            else:
                self.elements[key] = val


class Hamiltonian(object):
    def __init__(self, ham_dict):
        if type(ham_dict) == str:
            ham_dict = toml.load(ham_dict)
        self.name = ham_dict.get("name", "")

        self.nstypes = len(ham_dict["sites"])
        self.sites = [None for i in range(self.nstypes)]
        self.nitypes = len(ham_dict["interactions"])
        self.interactions = [None for i in range(self.nitypes)]

        for site in ham_dict["sites"]:
            S = Site(site)
            self.sites[S.id] = S
        for inter in ham_dict["interactions"]:
            Int = Interaction(inter)
            self.interactions[Int.id] = Int

        self.nxmax = 0
        for site in self.sites:
            self.nxmax = max(site.N, self.nxmax)

    def to_dict(self):
        return {
            "name": self.name,
            "sites": list(map(lambda x: x.to_dict(), self.sites)),
            "interactions": list(map(lambda x: x.to_dict(), self.interactions)),
        }

    def write_toml(self, filename):
        with codecs.open(filename, "w", "utf_8") as f:
            toml.dump(self.to_dict(), f)


class GraphedHamiltonian(Hamiltonian):
    def __init__(self, ham, lat):
        if not isinstance(ham, Hamiltonian):
            super(GraphedHamiltonian, self).__init__(ham)
        else:
            super(GraphedHamiltonian, self).__init__(ham.to_dict())
        self.load_lattice(lat)

    def load_lattice(self, lat):
        if type(lat) == str:
            lat = Lattice(lat)
        self.indeed_interactions = [
            IndeedInteraction(self.sites, self.interactions, v) for v in lat.vertices
        ]
        self.nitypes = len(self.indeed_interactions)

    def write_xml(self, filename):
        with codecs.open(filename, "w", "utf_8") as f:

            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write("<Hamiltonian>\n")
            indent = "  "
            level = 1

            f.write(indent * level + "<General>\n")
            level += 1
            f.write(indent * level + tagged("Comment", self.name))
            f.write(indent * level + tagged("NSTYPE", self.nstypes))
            f.write(indent * level + tagged("NITYPE", self.nitypes))
            f.write(indent * level + tagged("NXMAX", self.nxmax))
            level -= 1
            f.write(indent * level + "</General>\n")
            f.write("\n")

            for site in self.sites:
                f.write(indent * level + "<Site>\n")
                level += 1
                f.write(indent * level + tagged("STYPE", site.id))
                f.write(indent * level + tagged("TTYPE", 0))
                f.write(indent * level + tagged("NX", site.N))
                level -= 1
                f.write(indent * level + "</Site>\n")
                f.write("\n")

            for site in self.sites:
                f.write(indent * level + "<Source>\n")
                level += 1
                f.write(indent * level + tagged("STYPE", site.id))
                f.write(indent * level + tagged("TTYPE", 0))
                for elem in site.sources:
                    if site.sources[elem] == 0.0:
                        continue
                    f.write(
                        indent * level
                        + tagged(
                            "Weight", [elem[0][0], elem[1][0], abs(site.sources[elem])]
                        )
                    )
                level -= 1
                f.write(indent * level + "</Source>\n")
                f.write("\n")

            for interaction in self.indeed_interactions:
                nbody = interaction.nbody
                f.write(indent * level + "<Interaction>\n")
                level += 1
                f.write(indent * level + tagged("ITYPE", interaction.itype))
                f.write(indent * level + tagged("NBODY", nbody))
                f.write(indent * level + tagged("STYPE", interaction.stypes))

                for elem, v in interaction.elements.items():
                    v = -v
                    if v == 0.0:
                        continue
                    f.write(indent * level + "<Weight> ")
                    for i, j in zip(elem[0], elem[1]):
                        f.write("{0} {1} ".format(i, j))
                    if elem[0] != elem[1]:
                        v = abs(v)
                    f.write("{0:0< 18} ".format(v))
                    f.write("</Weight>\n")

                level -= 1
                f.write(indent * level + "</Interaction>\n")
                f.write("\n")

            level -= 1
            f.write("</Hamiltonian>\n")



#!python

# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import codecs

from .algorithm import Algorithm
from .displacement import Displacement
from .hamiltonian import GraphedHamiltonian
from .lattice import Lattice
from .parameter import Parameter
from .prob_kernel import (heat_bath, metropolis, reversible_suwa_todo,
                          suwa_todo)
from .std_lattice import std_lattice
from .std_model import std_model
from .util import ERROR, INFO
from .wavevector import Wavevector


def dla_pre(param, pfile):

    p = Parameter(param)

    lat = std_lattice(param)
    lat.write_xml(p["latfile"])

    h = std_model(param["hamiltonian"])
    ham = GraphedHamiltonian(h, lat)

    palg = param.get("algorithm", {})
    extra_shift = palg.get("extra_shift", 0.0)
    kernel = palg.get("kernel", "suwa todo")

    if kernel == "suwa todo":
        kernel = suwa_todo
    elif kernel == "reversible suwa todo":
        kernel = reversible_suwa_todo
    elif kernel == "heat bath":
        kernel = heat_bath
    elif kernel == "metropolis":
        kernel = metropolis
    elif args.kernel == "metropolice":
        ERROR('kenel = "metropolice" is now an invalid option because this is a typographic error in old DSQSS. Use "metropolis".')
    else:
        ERROR("unknown kernel: {0}".format(kernel))

    alg = Algorithm(ham, prob_kernel=kernel, ebase_extra=extra_shift)
    alg.write_xml(p["algfile"])

    if p["wvfile"] != "":
        wv = Wavevector()
        wv.generate(param.get("kpoints", {}), size=lat.size)
        wv.write_xml(p["wvfile"], lat)

    if p["dispfile"] != "":
        pdisp = param.get("displacement", {})
        disp = Displacement(lat, pdisp.get("distance_only", False))
        disp.write_xml(p["dispfile"])

    p.write_param(pfile)


def main():
    import argparse
    import sys
    import toml

    parser = argparse.ArgumentParser(
        description="Generate input files for dsqss/dla", add_help=True
    )

    parser.add_argument(
        "input",
        nargs="?",
        default=sys.stdin,
        type=argparse.FileType("r"),
        help="Input TOML file",
    )
    parser.add_argument(
        "-p", "--paramfile", dest="pfile", default="param.in", help="Parameter file"
    )

    args = parser.parse_args()
    if args.input is sys.stdin:
        INFO("waiting for standard input...")

    dla_pre(toml.load(args.input), args.pfile)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import sys
from copy import deepcopy


def INFO(msg, linebreak=True, file=sys.stderr):
    print("INFO: {0}\n".format(msg), file=file)
    if linebreak:
        print("\n", file=file)


def WARN(msg, linebreak=True, file=sys.stderr):
    print("WARN: {0}\n".format(msg), file=file)
    if linebreak:
        print("\n", file=file)


def ERROR(msg, to_be_continued=False, linebreak=True, file=sys.stderr):
    print("ERROR: {0}".format(msg), file=file)
    if linebreak:
        print("\n", file=file)
    if not to_be_continued:
        sys.exit(1)


def extend_list(lst, N):
    last = lst[-1]
    for i in range(N - len(lst)):
        lst.append(deepcopy(last))
    return lst


def get_as_list(d, name, default=None, extendto=None):
    if default is None:
        v = d[name]
    else:
        v = d.get(name, default)
    if type(v) is not list:
        v = [v]
    if extendto is not None:
        extend_list(v, extendto)
    return v


def index2coord(index, size):
    D = len(size)
    i = 0
    r = [0 for d in range(D)]
    while index > 0:
        r[i] = index % size[i]
        index //= size[i]
        i += 1
    return r


def coord2index(r, size):
    index = 0
    block = 1
    for x, L in zip(r, size):
        index += block * x
        block *= L
    return index


def tagged(tagname, values):
    if not hasattr(type(values), "__iter__") or type(values) is str:
        values = [values]
    s = "<{0}>".format(tagname)
    for v in values:
        if type(v) is float:
            s += " {0:0< 18}".format(v)
        else:
            s += " {0}".format(v)
    s += " </{0}>\n".format(tagname)
    return s



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys


def read_keyvalues(inputfile):
    """
    read and eval `inputfile` and return a dictionary
    """
    if type(inputfile) is str:
        with open(inputfile) as f:
            return read_keyvalues(f)
    if inputfile is sys.stdin:
        print("Waiting for standard input...")
    data_list = inputfile.readlines()
    res = {}
    # get key and value
    for data in data_list:
        data.strip()
        d_re = re.search("(.*)#(.*)", data)
        if d_re is not None:
            data = d_re.group(1)
        pattern = "(.*)=(.*)"
        if re.search(pattern, data) is not None:
            d_re = re.search(pattern, data)
            key = d_re.group(1).strip().lower()
            val = d_re.group(2).strip().split(",")
            if len(val) == 1:
                val = val[0]
            res[key] = val
    return res


def parse_list(lst, N, typ=None):
    if type(lst) is not list:
        ret = [lst]
    else:
        ret = lst[:]
    last = ret[-1]
    for n in range(N - len(ret)):
        ret.append(last)
    if typ is not None:
        ret = list(map(typ, ret))
    return ret



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.special import cosdg, sindg

from ..util import get_as_list


def generate(param):
    dim = 2
    L = get_as_list(param, "L", extendto=dim)
    bc = get_as_list(param, "bc", default=True, extendto=dim)
    basis = np.array([[1.0, 0.0], [cosdg(120), sindg(120)]])

    sites = [
            {"siteid": 0, "type": 0, "coord": [0.0, 0.0]},
            {"siteid": 1, "type": 0, "coord": [0.5, 0.0]},
            {"siteid": 2, "type": 0, "coord": [0.5, 0.5]},
            ]
    bonds = [
        {
            "bondid": 0,
            "type": 0,
            "source": {"siteid": 0},
            "target": {"siteid": 1, "offset": [0, 0]},
        },
        {
            "bondid": 1,
            "type": 0,
            "source": {"siteid": 0},
            "target": {"siteid": 2, "offset": [0, 0]},
        },
        {
            "bondid": 2,
            "type": 0,
            "source": {"siteid": 1},
            "target": {"siteid": 2, "offset": [0, 0]},
        },
        {
            "bondid": 3,
            "type": 0,
            "source": {"siteid": 1},
            "target": {"siteid": 0, "offset": [1, 0]},
        },
        {
            "bondid": 4,
            "type": 0,
            "source": {"siteid": 2},
            "target": {"siteid": 1, "offset": [0, 1]},
        },
        {
            "bondid": 5,
            "type": 0,
            "source": {"siteid": 2},
            "target": {"siteid": 0, "offset": [1, 1]},
        },
    ]

    parameter = {"name": "kagome", "dim": dim, "L": L, "bc": bc, "basis": basis}
    unitcell = {"sites": sites, "bonds": bonds}

    return {"parameter": parameter, "unitcell": unitcell}



#!python

# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import codecs
from copy import deepcopy

from .util import ERROR, INFO


def set_default_values(param):
    for name, val in (
        ("npre", 1000),
        ("ntherm", 1000),
        ("ndecor", 1000),
        ("nmcs", 1000),
        ("nset", 10),
        ("simulationtime", 0.0),
        ("ntau", 10),
        ("seed", 198212240),
        ("nvermax", 100000),
        ("nsegmax", 100000),
        ("algfile", "algorithm.xml"),
        ("latfile", "lattice.xml"),
        ("wvfile", ""),
        ("dispfile", ""),
        ("outfile", "sample.log"),
        ("sfoutfile", "sf.dat"),
        ("cfoutfile", "cf.dat"),
        ("ckoutfile", "ck.dat"),
    ):
        if name not in param:
            param[name] = val


def check_mandatories(param):
    for name in ("beta",):
        if name not in param:
            ERROR("parameter {0} is not specified.".format(name))


class Parameter(dict):
    def __init__(self, param):
        super(Parameter, self).__init__()
        if "parameter" in param:
            p = deepcopy(param["parameter"])
        else:
            p = deepcopy(param)
        set_default_values(p)
        check_mandatories(p)
        self.update(p)

    def write_param(self, pfile):
        with codecs.open(pfile, "w", "utf-8") as f:
            for key in sorted(self.keys()):
                f.write("{0} = {1}\n".format(key, self.get(key)))


def main():
    import argparse
    import sys
    import toml

    parser = argparse.ArgumentParser(
        description="Generate input files for dsqss/dla", add_help=True
    )

    parser.add_argument(
        "input",
        nargs="?",
        default=sys.stdin,
        type=argparse.FileType("r"),
        help="Input TOML file",
    )
    parser.add_argument(
        "-o", "--output", dest="pfile", default="param.in", help="Parameter file"
    )

    args = parser.parse_args()
    if args.input is sys.stdin:
        INFO("waiting for standard input...")

    p = Parameter(toml.load(args.input))
    p.write_param(args.pfile)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import math
from collections import namedtuple
from copy import deepcopy
from itertools import chain, product

import numpy as np

from .hamiltonian import append_matelem, keystate
from .prob_kernel import suwa_todo
from .util import tagged

Channel = namedtuple("Channel", ["outleg", "state", "prob"])
SiteInitialConfiguration = namedtuple("SiteInitialConfiguration", ["state", "channels"])

InitialConfiguration = namedtuple(
    "InitialConfiguration", ["istate", "fstate", "direction", "newstate", "channels"]
)


def num_channels(sources, nx):
    nchs = np.zeros(nx, dtype=int)
    for source in sources:
        st = source[0][0]
        nchs[st] += 1
    return nchs


class AlgSite:
    def __init__(self, hamsite, cutoff=1e-10):
        self.id = hamsite.id
        self.N = hamsite.N
        self.values = hamsite.values
        self.vtype = self.id
        self.initialconfigurations = []
        for st in range(self.N):
            channels = []
            bouncew = 1.0
            for source in hamsite.sources:
                if source[0][0] != st:
                    continue
                weight = hamsite.sources[source] ** 2
                t = source[1][0]
                bouncew -= weight
                channels.append(Channel(0, t, 0.5 * weight))
                channels.append(Channel(1, t, 0.5 * weight))
            if bouncew > cutoff:
                channels.append(Channel(-1, -1, bouncew))
            self.initialconfigurations.append(SiteInitialConfiguration(st, channels))
        self.vertex = WormVertex(hamsite)

    def write_xml(self, f, level):
        indent = "  "
        f.write(indent * level + "<Site>\n")
        level += 1

        f.write(indent * level + tagged("STYPE", self.id))
        f.write(indent * level + tagged("NumberOfStates", self.N))
        f.write(indent * level + tagged("LocalStates", self.values))
        f.write(indent * level + tagged("VertexTypeOfSource", self.vtype))

        for ic in self.initialconfigurations:
            f.write("\n")
            f.write(indent * level + "<InitialConfiguration>\n")
            level += 1
            f.write(indent * level + tagged("State", ic.state))
            f.write(indent * level + tagged("NumberOfChannels", len(ic.channels)))
            for c in ic.channels:
                f.write(indent * level + "<Channel>")
                f.write(" {0} {1} {2:0< 18}".format(c.outleg, c.state, c.prob))
                f.write(" </Channel>\n")
            level -= 1
            f.write(indent * level + "</InitialConfiguration>\n")

        level -= 1
        f.write(indent * level + "</Site>\n")


class Vertex(object):
    def __init__(self, vtype, category, nbody, ics):
        self.id = vtype
        self.category = category
        self.nbody = nbody
        self.initialconfigurations = ics

    def write_xml(self, f, level):
        indent = "  "
        f.write(indent * level + "<Vertex>\n")
        level += 1
        f.write(indent * level + tagged("VTYPE", self.id))
        f.write(indent * level + tagged("VCATEGORY", self.category))
        f.write(indent * level + tagged("NBODY", self.nbody))
        f.write(
            indent * level
            + tagged("NumberOfInitialConfigurations", len(self.initialconfigurations))
        )
        for ic in self.initialconfigurations:
            f.write("\n")
            f.write(indent * level + "<InitialConfiguration>\n")
            level += 1

            f.write(indent * level + "<State>")
            for i, final in zip(ic.istate, ic.fstate):
                f.write(" {0} {1}".format(i, final))
            f.write(" </State>\n")

            f.write(indent * level + tagged("IncomingDirection", ic.direction))
            f.write(indent * level + tagged("NewState", ic.newstate))
            f.write(indent * level + tagged("NumberOfChannels", len(ic.channels)))

            for c in ic.channels:
                f.write(indent * level + tagged("Channel", (c.outleg, c.state, c.prob)))

            level -= 1
            f.write(indent * level + "</InitialConfiguration>\n")
        level -= 1
        f.write(indent * level + "</Vertex>\n")


class WormVertex(Vertex):
    def __init__(self, hamsite):
        self.id = hamsite.id
        self.category = 1
        self.nbody = 1
        self.initialconfigurations = []
        for source in hamsite.sources:
            self.initialconfigurations.append(
                InitialConfiguration(
                    source[0], source[1], 0, source[1][0], [Channel(-1, -1, 1.0)]
                )
            )
            self.initialconfigurations.append(
                InitialConfiguration(
                    source[0], source[1], 1, source[0][0], [Channel(-1, -1, 1.0)]
                )
            )


class IntVertex(Vertex):
    def __init__(self, vtype, nbody, ics):
        self.id = vtype
        self.category = 2
        self.nbody = nbody
        self.initialconfigurations = ics


class AlgInteraction:
    def __init__(self, hamint, hamsites, prob_kernel=suwa_todo, ebase_extra=0.0):
        self.itype = hamint.itype
        self.vtype = hamint.itype + len(hamsites)
        self.nbody = hamint.nbody
        self.prob_kernel = prob_kernel

        self.hamint = hamint
        self.sites = [hamsites[stype] for stype in hamint.stypes]
        self.sitesources = [deepcopy(site.sources) for site in self.sites]
        self.intelements = deepcopy(hamint.elements)
        self.ebase_negsign = float("inf")

        self.signs = {}

        sumw = 0.0
        maxd = -float("inf")
        maxo = -float("inf")
        for elem, v in self.intelements.items():
            v = -v
            if elem[0] == elem[1]:
                self.intelements[elem] = v
                self.ebase_negsign = min(self.ebase_negsign, v)
                maxd = max(maxd, v)
            else:
                append_matelem(
                    self.signs, istate=elem[0], fstate=elem[1], value=np.sign(v)
                )
                v = abs(v)
                self.intelements[elem] = v
                sumw += v
                maxo = max(maxo, v)
        self.ebase_negsign *= -1.0
        maxd += self.ebase_negsign

        ndiag = 0
        for st in product(*map(lambda site: range(site.N), self.sites)):
            ndiag += 1
            k = (st, st)
            if k in self.intelements:
                self.intelements[k] += self.ebase_negsign
                sumw += self.intelements[k]
            else:
                append_matelem(self.intelements, state=st, value=self.ebase_negsign)
                sumw += self.intelements[k]
        self.ebase_nobounce = max((2 * maxo - sumw) / ndiag, 0.0)

        self.ebase_extra = ebase_extra

        if self.ebase_extra + maxd + self.ebase_nobounce == 0.0:
            self.ebase_extra = 0.5 * maxo

        for st in product(*map(lambda site: range(site.N), self.sites)):
            k = (st, st)
            self.intelements[k] += self.ebase_nobounce + self.ebase_extra

        self.intelements = {k: v for k, v in self.intelements.items() if v > 0.0}

        initialconfigurations = []
        for st in self.intelements:
            for inleg in range(2 * self.nbody):
                insite = inleg // 2
                for instate in range(self.sites[insite].N):
                    ic = self.make_initialconfiguration(st, inleg, instate)
                    if not ic:
                        continue
                    else:
                        initialconfigurations.append(ic)
        self.vertex = IntVertex(self.vtype, self.nbody, initialconfigurations)

    def make_initialconfiguration(self, states, inleg, instate):
        nbody = self.nbody
        insite = inleg // 2
        intau = inleg % 2
        oldinstate = states[intau][insite]

        if intau == 0:
            p = self.sitesources[insite].get(
                keystate(istate=instate, fstate=states[0][insite]), 0.0
            )
        else:
            p = self.sitesources[insite].get(
                keystate(istate=states[1][insite], fstate=instate), 0.0
            )
        if p == 0.0:
            return False

        midstates = [list(states[0]), list(states[1])]
        midstates[intau][insite] = instate

        probs = []
        outlegs = []
        outstates = []
        # import pdb; pdb.set_trace()
        for outleg in range(2 * nbody):
            outsite = outleg // 2
            outtau = outleg % 2
            for outstate in range(self.sites[outsite].N):
                st = deepcopy(midstates)
                oldoutstate = st[outtau][outsite]
                st[outtau][outsite] = outstate
                p = self.intelements.get(keystate(istate=st[0], fstate=st[1]), 0.0)
                if outtau == 0:
                    p *= self.sitesources[outsite].get(
                        keystate(istate=oldoutstate, fstate=outstate), 0.0
                    )
                else:
                    p *= self.sitesources[outsite].get(
                        keystate(istate=outstate, fstate=oldoutstate), 0.0
                    )
                if p == 0.0:
                    continue
                probs.append(p)
                outlegs.append(outleg)
                outstates.append(outstate)
                if inleg == outleg and outstate == oldinstate:
                    incomeindex = len(probs) - 1
        if len(probs) == 0:
            return False
        W = self.prob_kernel(probs)[incomeindex, :]
        channels = [
            Channel(outleg, outstate, p)
            for outleg, outstate, p in zip(outlegs, outstates, W)
            if p > 0.0
        ]
        return InitialConfiguration(states[0], states[1], inleg, instate, channels)

    def write_xml(self, f, level):
        indent = "  "
        f.write(indent * level + "<Interaction>\n")
        level += 1
        f.write(indent * level + tagged("ITYPE", self.itype))
        f.write(indent * level + tagged("VTYPE", self.vtype))
        f.write(indent * level + tagged("NBODY", self.nbody))
        f.write(
            indent * level
            + tagged(
                "EBASE", self.ebase_negsign + self.ebase_nobounce + self.ebase_extra
            )
        )
        for st in self.intelements:
            if st[0] != st[1]:
                continue
            f.write(
                indent * level
                + tagged("VertexDensity", chain(st[0], [self.intelements[st]]))
            )
        for st, v in self.signs.items():
            f.write(indent * level + "<Sign> ")
            for ini, fin in zip(st[0], st[1]):
                f.write("{0} {1} ".format(ini, fin))
            f.write("{0} </Sign>\n".format(v))
        level -= 1
        f.write(indent * level + "</Interaction>\n")


class Algorithm:
    def __init__(self, ham, prob_kernel=suwa_todo, ebase_extra=0.0):
        self.name = ham.name
        self.nstypes = ham.nstypes
        self.nitypes = ham.nitypes
        self.nxmax = ham.nxmax
        hamsites = [deepcopy(hamsite) for hamsite in ham.sites]
        for site in hamsites:
            maxprob = 0.0
            nchs = num_channels(site.sources, site.N)
            for source in site.sources:
                st = source[0][0]
                maxprob = max(maxprob, nchs[st] * site.sources[source] ** 2)
            for source in site.sources:
                site.sources[source] /= math.sqrt(maxprob)

        self.sites = [AlgSite(site) for site in hamsites]
        self.interactions = [
            AlgInteraction(
                hamint, hamsites, prob_kernel=prob_kernel, ebase_extra=ebase_extra
            )
            for hamint in ham.indeed_interactions
        ]

    def write_xml(self, filename):
        with codecs.open(filename, "w", "utf_8") as f:
            indent = "  "
            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write("<Algorithm>\n")
            level = 1

            f.write(level * indent + tagged("Comment", self.name))
            f.write(level * indent + "<General>\n")
            level += 1
            f.write(level * indent + tagged("NSTYPE", self.nstypes))
            f.write(level * indent + tagged("NITYPE", self.nitypes))
            f.write(level * indent + tagged("NVTYPE", self.nstypes + self.nitypes))
            f.write(level * indent + tagged("NXMAX", self.nxmax))
            f.write(level * indent + tagged("WDIAG", 0.25))
            level -= 1
            f.write(level * indent + "</General>\n")
            f.write("\n")

            for site in self.sites:
                site.write_xml(f, level)
            f.write("\n")
            for inter in self.interactions:
                inter.write_xml(f, level)
            f.write("\n")
            for site in self.sites:
                site.vertex.write_xml(f, level)
            f.write("\n")
            for inter in self.interactions:
                inter.vertex.write_xml(f, level)
            f.write("\n")

            f.write("</Algorithm>\n")



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
from itertools import product

import numpy as np

from .util import tagged


class Displacement:
    def __init__(self, lat, distance_only=False):
        self.lat = lat
        self.nr = 0
        sources = range(lat.nsites)
        self.displacements = -np.ones((lat.nsites, lat.nsites), dtype=int)
        self.nelems = lat.nsites * lat.nsites
        rdict = {}
        for s in sources:
            sr = np.array(lat.sites[s].coord)
            for t in range(lat.nsites):
                dr = np.array(lat.sites[t].coord) - sr
                for d in range(lat.dim):
                    if lat.bc[d] == 1 and dr[d] <= -lat.size[d] // 2:
                        dr[d] += lat.size[d]
                    elif lat.bc[d] == 1 and dr[d] > lat.size[d] // 2:
                        dr[d] -= lat.size[d]
                if distance_only:
                    r = sum(map(lambda x: x * x, np.dot(lat.latvec, dr)))
                else:
                    r = tuple(dr)
                if r not in rdict:
                    rdict[r] = self.nr
                    self.nr += 1
                self.displacements[s, t] = rdict[r]

    def write_xml(self, filename):
        lat = self.lat
        with codecs.open(filename, "w", "utf-8") as f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write("<Displacements>\n")
            f.write(tagged("Comment", lat.name))
            f.write(tagged("NumberOfKinds", self.nr))
            f.write(tagged("NumberOfSites", lat.nsites))
            f.write("\n")
            f.write("<!-- <R> [kind] [isite] [jsite] </R> -->\n")
            f.write("\n")

            for s, t in product(range(lat.nsites), range(lat.nsites)):
                f.write(tagged("R",
                               (self.displacements[s, t], s, t)))

            f.write("</Displacements>\n")



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy

import numpy as np


def heat_bath(weights, cutoff=1e-10):
    """
    return an array W,
    where W[i,j] is the probability of a transition i=>j
    calculated by the heat bath algorithm
    """
    ws = np.array(weights)
    ws /= ws.sum()
    return np.array([ws for i in range(len(ws))])


def metropolis(weights, cutoff=1e-10):
    """
    return an array W,
    where W[i,j] is the probability of a transition i=>j
    calculated by the Metropolice algorithm
    """
    N = len(weights)
    n = np.count_nonzero(weights)

    W = np.zeros((N, N))
    for i in range(N):
        if weights[i] == 0.0:
            continue
        acc = 0.0
        for j in range(N):
            if i == j:
                continue
            p = min(weights[j] / weights[i], 1.0) / (n - 1)
            if p < cutoff:
                p = 0.0
            W[i, j] = p
            acc += p
        W[i, i] = 1.0 - acc
    return W


def suwa_todo(weights, cutoff=1e-10):
    """
    return an array W,
    where W[i,j] is the probability of a transition i=>j
    calculated by the Suwa-Todo algorithm
    (H. Suwa and S. Todo, Phys. Rev. Lett. 105, 120603 (2010))
    """
    N = len(weights)
    W = np.zeros((N, N))
    indices = np.argsort(weights, kind="mergesort")[::-1]
    target = deepcopy(weights)
    for i in range(N):
        s = weights[indices[i]]
        j = 1
        while target[indices[j]] == 0.0:
            j = (j + 1) % N
        while s > 0.0:
            p = min(s, target[indices[j]])
            s -= p
            target[indices[j]] -= p
            W[indices[i], indices[j]] = p
            if target[indices[j]] <= 0.0:
                target[indices[j]] = 0.0
                if j == i:
                    break
                j = (j + 1) % N
    W /= W.sum(axis=1).reshape(-1, 1)
    W[W < cutoff] = 0.0
    return W / W.sum(axis=1).reshape(-1, 1)


def reversible_suwa_todo(weights, cutoff=1e-10):
    """
    return an array W,
    where W[i,j] is the probability of a transition i=>j
    calculated by the Suwa-Todo algorithm
    under the detailed balance condition
    (H. Suwa and S. Todo, arXiv:1106.3562)
    """
    N = len(weights)
    indices = np.argsort(weights, kind="mergesort")[::-1]
    weights = np.array(weights)
    W = np.zeros([N, N])
    for i in range(N):
        I = indices[i]
        W[I, I] = weights[I]

    S3 = sum(weights[indices[2:]])
    wdiff = weights[indices[0]] - weights[indices[1]]
    if wdiff >= S3:
        for i in range(1, N):
            I = indices[i]
            rst_swap(indices[0], I, weights[I], W)
    else:
        w3 = wdiff / S3
        for i in range(2, N):
            I = indices[i]
            v = w3 * weights[I]
            rst_swap(0, I, v, W)
        for j in range(N - 1, 0, -1):
            J = indices[j]
            v = W[J, J] / j
            for k in range(j - 1, -1, -1):
                rst_swap(J, indices[k], v, W)
    W /= W.sum(axis=1).reshape(-1, 1)
    W[W < cutoff] = 0.0
    return W / W.sum(axis=1).reshape(-1, 1)


def rst_swap(i, j, w, W):
    W[i, i] -= w
    W[i, j] += w
    W[j, i] += w
    W[j, j] -= w



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

setup(
    name="dsqss",
    version="2.0.4",
    description="DSQSS input files generator",
    url="https://github.com/issp-center-dev/dsqss",
    author="DSQSS developers",
    author_email="dsqss-dev@issp.u-tokyo.ac.jp",
    license="GPLv3",
    packages=["dsqss", 'dsqss.lattice_factory'],
    python_requires=">=2.7",
    install_requires=["numpy", "scipy", "toml"],
    entry_points={
        "console_scripts": [
            "dla_hamgen = dsqss.std_model:main",
            "dla_latgen = dsqss.std_lattice:main",
            "dla_pgen = dsqss.parameter:main",
            "dla_wvgen = dsqss.wavevector:main",
            "dla_alg = dsqss.dla_alg:main",
            "dla_pre = dsqss.dla_pre:main",
            "pmwa_pre = dsqss.pmwa_pre:main",
        ]
    },
    zip_safe=False,
)



#!python

# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import argparse

from .algorithm import Algorithm
from .displacement import Displacement
from .hamiltonian import GraphedHamiltonian
from .lattice import Lattice
from .prob_kernel import (heat_bath, metropolis, reversible_suwa_todo,
                          suwa_todo)
from .util import ERROR
from .wavevector import Wavevector


def dla_alg(
    lat,
    ham=None,
    wv=None,
    latxml=None,
    algxml=None,
    wvxml=None,
    dispxml=None,
    prob_kernel=suwa_todo,
    ebase_extra=0.0,
    distance_only=False,
):
    if latxml is not None:
        lat.write_xml(latxml)
    if ham is not None and algxml is not None:
        alg = Algorithm(ham, prob_kernel=prob_kernel, ebase_extra=ebase_extra)
        alg.write_xml(algxml)
    if wv is not None and wvxml is not None:
        wv.write_xml(wvxml, lat)
    if dispxml is not None:
        disp = Displacement(lat, distance_only=distance_only)
        disp.write_xml(dispxml)


def main():

    parser = argparse.ArgumentParser(
        description="Generate algorithm and lattice XML files for DSQSS/DLA",
        conflict_handler="resolve",
        epilog=(
            "Note that the input lattice file is ALWAYS required "
            "even if the lattice XML file will not be generated "
            "(--without_lattice)."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=True,
    )

    parser.add_argument(
        "-l", "--lattice", dest="lat", default="lattice.dat", help="Input Lattice file"
    )
    parser.add_argument(
        "-h",
        "--hamiltonian",
        dest="ham",
        default="hamiltonian.toml",
        help="Input Hamiltonian file",
    )
    parser.add_argument(
        "-L",
        "--LatticeXML",
        dest="latxml",
        default="lattice.xml",
        help="Output Lattice XML file",
    )
    parser.add_argument(
        "-A",
        "--AlgorithmXML",
        dest="algxml",
        default="algorithm.xml",
        help="Output Algorithm XML file",
    )
    parser.add_argument(
        "--without_lattice",
        dest="wolat",
        action="store_true",
        help="Never output Lattice XML file",
    )
    parser.add_argument(
        "--without_algorithm",
        dest="woalg",
        action="store_true",
        help="Never output Algorithm XML file",
    )
    parser.add_argument(
        "-k", "--kpoint", dest="kpoint", default=None, help="kpoints data file"
    )
    parser.add_argument("--wv", dest="wv", default="wv.xml", help="wavevector XML file")
    parser.add_argument(
        "--disp", dest="disp", default=None, help="displacement XML file"
    )
    parser.add_argument(
        "--distance-only",
        dest="distance_only",
        action="store_true",
        help="use |r_ij| instead of r_ij",
    )
    parser.add_argument(
        "--kernel",
        dest="kernel",
        default="suwa todo",
        help=(
            "transition kernel of Monte Carlo: "
            '"suwa todo", "reversible suwa todo", '
            '"heat bath", or "metropolis".'
        ),
    )
    parser.add_argument(
        "--extra-shift",
        dest="extra_shift",
        type=float,
        default=0.0,
        help="extra energy shift",
    )

    args = parser.parse_args()

    if args.kernel == "suwa todo":
        kernel = suwa_todo
    elif args.kernel == "reversible suwa todo":
        kernel = reversible_suwa_todo
    elif args.kernel == "heat bath":
        kernel = heat_bath
    elif args.kernel == "metropolis":
        kernel = metropolis
    elif args.kernel == "metropolice":
        ERROR('kenel = "metropolice" is now an invalid option because this is a typographic error in old DSQSS. Use "metropolis".')
    else:
        ERROR("unknown kernel: {0}".format(args.kernel))

    lat = Lattice(args.lat)
    ham = GraphedHamiltonian(args.ham, lat) if not args.woalg else None
    wv = None
    if args.kpoint is not None:
        wv = Wavevector()
        wv.load(args.kpoint)

    dla_alg(
        lat,
        ham=ham,
        wv=wv,
        latxml=args.latxml,
        algxml=args.algxml,
        wvxml=args.wv,
        dispxml=args.disp,
        prob_kernel=kernel,
        ebase_extra=args.extra_shift,
        distance_only=args.distance_only,
    )

    # lat = Lattice(args.lat)
    # if not args.wolat:
    #     lat.write_xml(args.latxml)
    # if not args.woalg:
    #     ham = GraphedHamiltonian(args.ham, lat)
    #     alg = Algorithm(ham, prob_kernel=kernel, ebase_extra=args.extra_shift)
    #     alg.write_xml(args.algxml)
    # if args.kpoint is not None:
    #     wv = Wavevector()
    #     wv.load(args.kpoint)
    #     wv.write_xml(args.wv, lat)
    # if args.disp is not None:
    #     disp = Displacement(lat, distance_only=args.distance_only)
    #     disp.write_xml(args.disp)


if __name__ == "__main__":
    main()



# DSQSS (Discrete Space Quantum Systems Solver)
# Copyright (C) 2018- The University of Tokyo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version. 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..util import get_as_list


def generate(param):
    dim = param["dim"]
    L = get_as_list(param, "L", extendto=dim)
    bc = get_as_list(param, "bc", default=True, extendto=dim)
    basis = np.eye(dim)

    sites = [{"siteid": 0, "type": 0, "coord": [0.0] * dim}]
    bonds = []
    for d in range(dim):
        offset = [0] * dim
        offset[d] = 1
        bonds.append(
            {
                "bondid": d,
                "type": 0,
                "source": {"siteid": 0},
                "target": {"siteid": 0, "offset": offset},
            }
        )

    parameter = {"name": "hypercubic", "dim": dim, "L": L, "bc": bc, "basis": basis}
    unitcell = {"sites": sites, "bonds": bonds}

    return {"parameter": parameter, "unitcell": unitcell}
